/* {{{ Copyright */

/*

SUSAN Version 2i_medx - nonlinear 2D/3D smoothing

Oxford Centre for Functional Magnetic Resonance Imaging of the Brain,
Department of Clinical Neurology, Oxford University, Oxford, UK
(Previously in Computer Vision and Image Processing Group - now
Computer Vision and Electro Optics Group - DERA Chertsey, UK)
Email:    steve@fmrib.ox.ac.uk
WWW:      http://www.fmrib.ox.ac.uk/~steve

(C) Crown Copyright (1995-1999), Defence Evaluation and Research Agency,
Farnborough, Hampshire, GU14 6TD, UK
DERA WWW site:
http://www.dera.gov.uk/
DERA Computer Vision and Electro Optics Group WWW site:
http://www.dera.gov.uk/imageprocessing/dera/group_home.html
DERA Computer Vision and Electro Optics Group point of contact:
Dr. John Savage, jtsavage@dera.gov.uk, +44 1344 633203

A UK patent has been granted: "Method for digitally processing
images to determine the position of edges and/or corners therein for
guidance of unmanned vehicle", UK Patent 2272285. Proprietor:
Secretary of State for Defence, UK. 15 January 1997

This code is issued for research purposes only and remains the
property of the UK Secretary of State for Defence. This code must
not be passed on without this header information being kept
intact. This code must not be sold.

*/

/* }}} */
/* {{{ Readme First */

/**********************************************************************\

This 3D version derived from 2d version SUSAN Version 2i

SUSAN = Smallest Univalue Segment Assimilating Nucleus

Email:    steve@fmrib.ox.ac.uk
WWW:      http://www.fmrib.ox.ac.uk/~steve

Related paper:
@article{Smith97,
author = "Smith, S.M. and Brady, J.M.",
title = "{SUSAN} - A New Approach to Low Level Image Processing",
journal = "Int. Journal of Computer Vision",
pages = "45--78",
volume = "23",
number = "1",
month = "May",
year = 1997}

To be registered for automatic (bug) updates of SUSAN, send an email.

Note: FDT is the image data type

Note: edge and corner finding have been taken out of the 3D version
of SUSAN - it is not thought that they would be of great value.

See following section for different machine information. Please
report any bugs (and fixes). There are a few optional changes that
can be made in the "defines" section which follows shortly.

This code is written using an emacs folding mode, making moving
around the different sections very easy. This is why there are
various marks within comments and why comments are indented.

SPATIAL CONTROL: d

In SUSAN smoothing d controls the size of the Gaussian mask; its
default is 4.0. Increasing d gives more smoothing. In edge finding,
a fixed flat mask is used, either 37 pixels arranged in a "circle"
(default), or a 3 by 3 mask which gives finer detail. In corner
finding, only the larger 37 pixel mask is used; d is not
variable. In smoothing, the flat 3 by 3 mask can be used instead of
a larger Gaussian mask; this gives low smoothing and fast operation.

BRIGHTNESS CONTROL: t

In all three algorithms, t can be varied (default=20); this is the
main threshold to be varied. It determines the maximum difference in
greylevels between two pixels which allows them to be considered
part of the same "region" in the image. Thus it can be reduced to
give more edges or corners, i.e. to be more sensitive, and vice
versa. In smoothing, reducing t gives less smoothing, and vice
versa. Set t=10 for the test image available from the SUSAN web
page.

ITERATIONS:

With SUSAN smoothing, more smoothing can also be obtained by
iterating the algorithm several times. This has a different effect
from varying d or t.

BRIGHTNESS FUNCTION LUT IMPLEMENTATION:
(Only read this if you are interested in the C implementation)

The SUSAN brightness function is implemented as a LUT
(Look-Up-Table) for speed. The resulting pointer-based code is a
little hard to follow, so here is a brief explanation. In
setup_brightness_lut() the LUT is setup. This mallocs enough space
for *bp and then repositions the pointer to the centre of the
malloced space. The SUSAN function e^-(x^6) or e^-(x^2) is
calculated and converted to a unsigned char in the range 0-100, for all
possible image brightness differences (including negative
ones). Thus bp[23] is the output for a brightness difference of 23
greylevels. In the SUSAN algorithms this LUT is used as follows:

p=in + (i-3)*x_size + j - 1;
p points to the first image pixel in the circular mask surrounding
point (x,y).

cp=bp + in[i*x_size+j];
cp points to a position in the LUT corresponding to the brightness
of the centre pixel (x,y).

now for every pixel within the mask surrounding (x,y),
n+=*(cp-*p++);
the brightness difference function is found by moving the cp pointer
down by an amount equal to the value of the pixel pointed to by p,
thus subtracting the two brightness values and performing the
exponential function. This value is added to n, the running USAN
area.

in SUSAN smoothing, the variable height mask is implemented by
multiplying the above by the moving mask pointer, reset for each new
centre pixel.
tmp = *dpt++ * *(cp-brightness);

\**********************************************************************/

//#include "stdafx.h"


//-----------JNI HEADER_____________

/* DO NOT EDIT THIS FILE - it is machine generated */
#include "jni.h"
/* Header for class SusanFilter3D */

#ifndef _Included_SusanFilter3D
#define _Included_SusanFilter3D
#ifdef __cplusplus
extern "C"
{
#endif
/*
 * Class:     SusanFilter3D
 * Method:    susan_smoothing3d
 * Signature: (Z[BFIIIIZ)V
 */
  JNIEXPORT void JNICALL Java_SusanFilter3D_susan_1smoothing3d
    (JNIEnv *, jclass, jboolean, jbyteArray, jfloat, jint, jint, jint, jint,
     jboolean);

/*
 * Class:     SusanFilter3D
 * Method:    susan_smoothing2d
 * Signature: (Z[BFIIIIZ)V
 */
  JNIEXPORT void JNICALL Java_SusanFilter3D_susan_1smoothing2d
    (JNIEnv *, jclass, jboolean, jbyteArray, jfloat, jint, jint, jint, jint,
     jboolean);

/*
* Class:     SusanFilter3D
* Method:    susan_edges
* Signature: (Z[BZZIIIII)V
*/
  JNIEXPORT void JNICALL Java_SusanFilter3D_susan_1edges
    (JNIEnv *, jclass, jboolean, jbyteArray, jboolean, jboolean, jint, jint,
     jint, jint, jint);

#ifdef __cplusplus
}
#endif
#endif

//-------------------------------

typedef int TOTAL_TYPE;

#include "libssbase.h"

#define SEVEN_SUPP		/* size for non-max corner suppression; SEVEN_SUPP or FIVE_SUPP */
#define MAX_CORNERS   15000	/* max corners per frame */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#define  exit_error(IFB,IFC) { fprintf(stderr,IFB,IFC); exit(0); }
#define  FTOI(a) ( (a) < 0 ? ((int)(a-0.5)) : ((int)(a+0.5)) )
typedef unsigned char uchar;
typedef struct
{
  int x, y, info, dx, dy, I;
} CORNER_LIST[MAX_CORNERS];

/* }}} */
/* {{{ usage() */

void
usage ()
{
  printf ("Usage: susan <in.pgm> <out.pgm> [options]\n\n");

  printf ("-s : Smoothing mode (default)\n");
  printf ("-e : Edges mode\n");
  printf ("-c : Corners mode\n\n");

  printf
    ("See source code for more information about setting the thresholds\n");
  printf ("-t <thresh> : Brightness threshold, all modes (default=20)\n");
  printf
    ("-d <thresh> : Distance threshold, smoothing mode, (default=4) (use next option instead for flat 3x3 mask)\n");
  printf ("-3 : Use flat 3x3 mask, edges or smoothing mode\n");
  printf
    ("-n : No post-processing on the binary edge map (runs much faster); edges mode\n");
  printf
    ("-q : Use faster (and usually stabler) corner mode; edge-like corner suppression not carried out; corners mode\n");
  printf
    ("-b : Mark corners/edges with single black points instead of black with white border; corners or edges mode\n");
  printf
    ("-p : Output initial enhancement image only; corners or edges mode (default is edges mode)\n");

  printf
    ("\nSUSAN Version 2l (C) 1995-1997 Stephen Smith, DRA UK. steve@fmrib.ox.ac.uk\n");

  exit (0);
}

/* }}} */
/* {{{ get_image(filename,in,x_size,y_size) */

/* {{{ int getint(fp) derived from XV */

/* }}} */


void
setProgress (JNIEnv * env, jclass susan, int num, int total)
{
  jmethodID mid = (env)->GetStaticMethodID (susan, "setProgress", "(II)V");
  if (mid == 0)
    {
      printf ("method not found \n");
      return;
    }
  (env)->CallStaticVoidMethod (susan, mid, num, total);
}


FDT
susan_median3d (FDT * in, int i, int j, int k, int x_size, int y_size)
{
  FDT p[26], tmp;
  int kk, ll;

  p[0] = in[(k - 1) * x_size * y_size + (i - 1) * x_size + j - 1];
  p[1] = in[(k - 1) * x_size * y_size + (i - 1) * x_size + j];
  p[2] = in[(k - 1) * x_size * y_size + (i - 1) * x_size + j + 1];
  p[3] = in[(k - 1) * x_size * y_size + (i) * x_size + j - 1];
  p[4] = in[(k - 1) * x_size * y_size + (i) * x_size + j];
  p[5] = in[(k - 1) * x_size * y_size + (i) * x_size + j + 1];
  p[6] = in[(k - 1) * x_size * y_size + (i + 1) * x_size + j - 1];
  p[7] = in[(k - 1) * x_size * y_size + (i + 1) * x_size + j];
  p[8] = in[(k - 1) * x_size * y_size + (i + 1) * x_size + j + 1];

  p[9] = in[k * x_size * y_size + (i - 1) * x_size + j - 1];
  p[10] = in[k * x_size * y_size + (i - 1) * x_size + j];
  p[11] = in[k * x_size * y_size + (i - 1) * x_size + j + 1];
  p[12] = in[k * x_size * y_size + (i) * x_size + j - 1];
  p[13] = in[k * x_size * y_size + (i) * x_size + j + 1];
  p[14] = in[k * x_size * y_size + (i + 1) * x_size + j - 1];
  p[15] = in[k * x_size * y_size + (i + 1) * x_size + j];
  p[16] = in[k * x_size * y_size + (i + 1) * x_size + j + 1];

  p[17] = in[(k + 1) * x_size * y_size + (i - 1) * x_size + j - 1];
  p[18] = in[(k + 1) * x_size * y_size + (i - 1) * x_size + j];
  p[19] = in[(k + 1) * x_size * y_size + (i - 1) * x_size + j + 1];
  p[20] = in[(k + 1) * x_size * y_size + (i) * x_size + j - 1];
  p[21] = in[(k + 1) * x_size * y_size + (i) * x_size + j];
  p[22] = in[(k + 1) * x_size * y_size + (i) * x_size + j + 1];
  p[23] = in[(k + 1) * x_size * y_size + (i + 1) * x_size + j - 1];
  p[24] = in[(k + 1) * x_size * y_size + (i + 1) * x_size + j];
  p[25] = in[(k + 1) * x_size * y_size + (i + 1) * x_size + j + 1];

  for (kk = 0; kk < 25; kk++)
    for (ll = 0; ll < (25 - kk); ll++)
      if (p[ll] > p[ll + 1])
	{
	  tmp = p[ll];
	  p[ll] = p[ll + 1];
	  p[ll + 1] = tmp;
	}

  return ((FDT) (((double) p[12] + (double) p[13]) / 2.0));
}


void
senlarge3d (FDT ** in, FDT * tmp_image, int *x_size, int *y_size, int *z_size,
	    int border)
{
  int i, j, k;

  for (k = 0; k < *z_size; k++)	/* copy *in into tmp_image */
    for (i = 0; i < *y_size; i++)
      memcpy (tmp_image +
	      (k + border) * (*x_size + 2 * border) * (*y_size + 2 * border) +
	      (i + border) * (*x_size + 2 * border) + border,
	      *in + k * *x_size * *y_size + i * *x_size,
	      *x_size * sizeof (FDT));

  for (k = 0; k < *z_size; k++)
    {
      for (i = 0; i < border; i++)	/* copy top and bottom rows; invert as many as necessary */
	{
	  memcpy (tmp_image +
		  (k + border) * (*x_size + 2 * border) * (*y_size +
							   2 * border) +
		  (border - 1 - i) * (*x_size + 2 * border) + border,
		  *in + k * *y_size * *x_size + i * *x_size,
		  *x_size * sizeof (FDT));
	  memcpy (tmp_image +
		  (k + border) * (*x_size + 2 * border) * (*y_size +
							   2 * border) +
		  (*y_size + border + i) * (*x_size + 2 * border) + border,
		  *in + k * *y_size * *x_size + (*y_size - i - 1) * *x_size,
		  *x_size * sizeof (FDT));
	}
      for (i = 0; i < border; i++)	/* copy left and right columns */
	for (j = 0; j < *y_size + 2 * border; j++)
	  {
	    tmp_image[(k + border) * (*x_size + 2 * border) * (*y_size +
							       2 * border) +
		      j * (*x_size + 2 * border) + border - 1 - i] =
	      tmp_image[(k + border) * (*x_size + 2 * border) * (*y_size +
								 2 * border) +
			j * (*x_size + 2 * border) + border + i];
	    tmp_image[(k + border) * (*x_size + 2 * border) * (*y_size +
							       2 * border) +
		      j * (*x_size + 2 * border) + *x_size + border + i] =
	      tmp_image[(k + border) * (*x_size + 2 * border) * (*y_size +
								 2 * border) +
			j * (*x_size + 2 * border) + *x_size + border - 1 -
			i];
	  }
    }

  for (k = 0; k < border; k++)
    {
      memcpy (tmp_image +
	      (border - 1 - k) * (*x_size + 2 * border) * (*y_size +
							   2 * border),
	      tmp_image + (border + k) * (*x_size + 2 * border) * (*y_size +
								   2 *
								   border),
	      (*x_size + 2 * border) * (*y_size + 2 * border) * sizeof (FDT));
      memcpy (tmp_image +
	      (*z_size + border + k) * (*x_size + 2 * border) * (*y_size +
								 2 * border),
	      tmp_image + (*z_size + border - 1 - k) * (*x_size +
							2 * border) *
	      (*y_size + 2 * border),
	      (*x_size + 2 * border) * (*y_size + 2 * border) * sizeof (FDT));
    }

  *x_size += 2 * border;	/* alter image size */
  *y_size += 2 * border;
  *z_size += 2 * border;
  *in = tmp_image;		/* repoint in */

  /* {{{ COMMENT output tmp_image to file for debugging */

#ifdef FoldingComment

  {
    FILE *ofp;
    char filename[100];

    sprintf (filename, "/tmp/Enlarge.raw%d", in);
    ofp = fopen (filename, "wb");
    fwrite (tmp_image, *x_size * *y_size * *z_size * sizeof (FDT), 1, ofp);
    fprintf (stderr, "%d %d %d", *x_size, *y_size, *z_size);
  }

#endif

  /* }}} */
}

void
senlarge (FDT ** in, FDT * tmp_image, int *x_size, int *y_size, int z,
	  int border)
{
  int i, j;

  for (i = 0; i < *y_size; i++)	/* copy *in into tmp_image */
    memcpy (tmp_image + (i + border) * (*x_size + 2 * border) + border,
	    *in + i * *x_size, *x_size * sizeof (FDT));

  for (i = 0; i < border; i++)	/* copy top and bottom rows; invert as many as necessary */
    {
      memcpy (tmp_image + (border - 1 - i) * (*x_size + 2 * border) + border,
	      *in + i * *x_size, *x_size * sizeof (FDT));
      memcpy (tmp_image + (*y_size + border + i) * (*x_size + 2 * border) +
	      border, *in + (*y_size - i - 1) * *x_size,
	      *x_size * sizeof (FDT));
    }

  for (i = 0; i < border; i++)	/* copy left and right columns */
    for (j = 0; j < *y_size + 2 * border; j++)
      {
	tmp_image[j * (*x_size + 2 * border) + border - 1 - i] =
	  tmp_image[j * (*x_size + 2 * border) + border + i];
	tmp_image[j * (*x_size + 2 * border) + *x_size + border + i] =
	  tmp_image[j * (*x_size + 2 * border) + *x_size + border - 1 - i];
      }

  *x_size += 2 * border;	/* alter image size */
  *y_size += 2 * border;
  *in = tmp_image;		/* repoint in */

  /* {{{ COMMENT output tmp_image to file for debugging */

#ifdef FoldingComment

  /* {{{ output tmp_image to file for debugging */

  {
    FILE *ofp;
    char ff[100];

    sprintf (ff, "/tmp/Enlarge%d.raw", z);
    ofp = fopen (ff, "wb");
    fwrite (tmp_image, *x_size * *y_size * sizeof (FDT), 1, ofp);
    fprintf (stderr, "%d %d %d", *x_size, *y_size, z);
  }

  /* }}} */

#endif

  /* }}} */
}


FDT
susan_median (FDT * in, int i, int j, int x_size)
{
  FDT p[8], tmp;
  int k, l;

  p[0] = in[(i - 1) * x_size + j - 1];
  p[1] = in[(i - 1) * x_size + j];
  p[2] = in[(i - 1) * x_size + j + 1];
  p[3] = in[(i) * x_size + j - 1];
  p[4] = in[(i) * x_size + j + 1];
  p[5] = in[(i + 1) * x_size + j - 1];
  p[6] = in[(i + 1) * x_size + j];
  p[7] = in[(i + 1) * x_size + j + 1];

  for (k = 0; k < 7; k++)
    for (l = 0; l < (7 - k); l++)
      if (p[l] > p[l + 1])
	{
	  tmp = p[l];
	  p[l] = p[l + 1];
	  p[l + 1] = tmp;
	}

  return ((FDT) ((((float) p[3]) + ((float) p[4])) / 2.0));
}

void
susan_smoothing (int three_by_three, FDT * in, float x_dt, float y_dt,
		 int x_size, int y_size, int z, unsigned char *bp,
		 int use_median)
{
  /* {{{ vars */

  int increment, mask_size, x_mask_size = 1, y_mask_size = 1,
    i, j, x, y, area, tmp;
  FDT *ip, *tmp_image, *out = in, centre, brightness;
  unsigned char *dp, *dpt, *cp;
  TOTAL_TYPE total;

  /* }}} */

  /* {{{ setup larger image and border sizes */

  if (three_by_three == 0)
    {
      x_mask_size = ((int) (2.0 * x_dt)) + 1;
      y_mask_size = ((int) (2.0 * y_dt)) + 1;
    }
  else
    y_mask_size = x_mask_size = 1;

  mask_size = MAX (x_mask_size, y_mask_size);

  tmp_image =
    (FDT *) malloc ((x_size + mask_size * 2) * (y_size + mask_size * 2) *
		    sizeof (FDT));
  senlarge (&in, tmp_image, &x_size, &y_size, z, mask_size);

  /* }}} */

  if (three_by_three == 0)
    {				/* large Gaussian masks */
      /* {{{ setup distance lut */

      increment = x_size - ((x_mask_size * 2) + 1);

      dp =
	(unsigned char *) malloc (((x_mask_size * 2) + 1) *
				  ((y_mask_size * 2) + 1));
      dpt = dp;
      x_dt = -(2 * x_dt * x_dt);
      y_dt = -(2 * y_dt * y_dt);

      for (i = -y_mask_size; i <= y_mask_size; i++)
	{
	  for (j = -x_mask_size; j <= x_mask_size; j++)
	    {
	      x =
		(int) (100.0 *
		       exp (((float) (i * i)) / y_dt +
			    ((float) (j * j)) / x_dt));
	      /*fprintf(stderr,"%d ",x); */
	      *dpt++ = (unsigned char) x;
	    }
	  /*fprintf(stderr,"\n"); */
	}

      /* }}} */
      /* {{{ main section */

      for (i = mask_size; i < y_size - mask_size; i++)	/* use mask_size as senlarge was isotropic */
	{
	  for (j = mask_size; j < x_size - mask_size; j++)
	    {
	      area = 0;
	      total = 0;
	      dpt = dp;
	      ip = in + ((i - y_mask_size) * x_size) + j - x_mask_size;
	      centre = in[i * x_size + j];
	      cp = bp + centre;
	      for (y = -y_mask_size; y <= y_mask_size; y++)
		{
		  for (x = -x_mask_size; x <= x_mask_size; x++)
		    {
		      brightness = *ip++;
		      tmp = *dpt++ * *(cp - brightness);
		      area += tmp;
		      total += tmp * brightness;
		    }
		  ip += increment;
		}

	      if (use_median)
		{
		  tmp = area - 10000;
		  if (tmp < 10000)
		    *out++ = susan_median (in, i, j, x_size);
		  else
		    *out++ = ((total - (centre * 10000)) / tmp);
		}
	      else
		*out++ = total / area;
	    }
	}

      /* }}} */
    }
  else
    {				/* 3x3 constant mask */
      /* {{{ main section */

      for (i = 1; i < y_size - 1; i++)
	{
	  for (j = 1; j < x_size - 1; j++)
	    {
	      area = 0;
	      total = 0;
	      ip = in + ((i - 1) * x_size) + j - 1;
	      centre = in[i * x_size + j];
	      cp = bp + centre;

	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      ip += x_size - 2;
	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      ip += x_size - 2;
	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;

	      if (use_median)
		{
		  tmp = area - 100;
		  if (tmp < 100)
		    *out++ = susan_median (in, i, j, x_size);
		  else
		    *out++ = (total - (centre * 100)) / tmp;
		}
	      else
		*out++ = total / area;
	    }
	}

      /* }}} */
    }
}

void
susan_smoothing3d (JNIEnv * env, jclass susan, int three_by_three,
		   FDT * in,
		   float x_dt,
		   float y_dt,
		   float z_dt,
		   int x_size,
		   int y_size, int z_size, unsigned char *bp, int use_median)
{
  /* {{{ vars */




  int increment, mask_size, x_mask_size = 1, y_mask_size = 1, z_mask_size = 1,
    i, j, k, x, y, z, area, tmp;
  FDT *ip, *tmp_image, *out = in, centre, brightness;
  unsigned char *dp, *dpt, *cp;
  TOTAL_TYPE total;

  /* }}} */

  /* {{{ setup larger image and border sizes */

  if (three_by_three == 0)
    {
      x_mask_size = ((int) (2.0 * x_dt)) + 1;
      y_mask_size = ((int) (2.0 * y_dt)) + 1;
      z_mask_size = ((int) (2.0 * z_dt)) + 1;
    }
  else
    z_mask_size = y_mask_size = x_mask_size = 1;

  mask_size = MAX (MAX (x_mask_size, y_mask_size), z_mask_size);

  tmp_image =
    (FDT *) malloc ((x_size + mask_size * 2) * (y_size + mask_size * 2) *
		    (z_size + mask_size * 2) * sizeof (FDT));
  senlarge3d (&in, tmp_image, &x_size, &y_size, &z_size, mask_size);

  /* }}} */

  if (three_by_three == 0)
    {				/* large Gaussian masks */
      /* {{{ setup distance lut */

      increment = x_size - (x_mask_size * 2 + 1);

      dp =
	(unsigned char *) malloc (((x_mask_size * 2) + 1) *
				  ((y_mask_size * 2) +
				   1) * ((z_mask_size * 2) + 1));
      dpt = dp;
      x_dt = -(2 * x_dt * x_dt);
      y_dt = -(2 * y_dt * y_dt);
      z_dt = -(2 * z_dt * z_dt);



      for (k = -z_mask_size; k <= z_mask_size; k++)
	for (i = -y_mask_size; i <= y_mask_size; i++)
	  {
	    for (j = -x_mask_size; j <= x_mask_size; j++)
	      {
		x =
		  (int) (100.0 *
			 exp (((float) (i * i)) / y_dt +
			      ((float) (j * j)) / x_dt +
			      ((float) (k * k)) / z_dt));
		/*fprintf(stderr,"%.4d ",x); */
		*dpt++ = (unsigned char) x;
	      }
	    /*fprintf(stderr,"\n"); */
	  }

      /* }}} */
      /* {{{ main section */

      for (k = mask_size; k < z_size - mask_size; k++)	/* use mask_size as senlarge was isotropic */
	{

	  setProgress (env, susan, k - mask_size, z_size);


	  for (i = mask_size; i < y_size - mask_size; i++)
	    {
	      for (j = mask_size; j < x_size - mask_size; j++)
		{
		  area = 0;
		  total = 0;
		  dpt = dp;
		  ip =
		    in + (k - z_mask_size) * x_size * y_size + (i -
								y_mask_size) *
		    x_size + j - x_mask_size;
		  centre = in[k * x_size * y_size + i * x_size + j];
		  cp = bp + centre;
		  for (z = -z_mask_size; z <= z_mask_size; z++)
		    {
		      for (y = -y_mask_size; y <= y_mask_size; y++)
			{
			  for (x = -x_mask_size; x <= x_mask_size; x++)
			    {
			      brightness = *ip++;
			      tmp = *dpt++ * *(cp - brightness);
			      area += tmp;
			      total += tmp * brightness;
			    }
			  ip += increment;
			}
		      ip += x_size * (y_size - (y_mask_size * 2 + 1));
		    }

		  if (use_median)
		    {
		      tmp = area - 10000;
		      if (tmp < 20000)
			*out++ = susan_median3d (in, i, j, k, x_size, y_size);
		      else
			*out++ = ((total - (centre * 10000)) / tmp);
		    }
		  else
		    *out++ = total / area;
		}
	    }
	}

      /* }}} */
    }
  else
    {				/* 3x3 constant mask */
      /* {{{ main section */

      for (k = 1; k < z_size - 1; k++)
	{


	  setProgress (env, susan, k, z_size);



	  for (i = 1; i < y_size - 1; i++)
	    {
	      for (j = 1; j < x_size - 1; j++)
		{
		  area = 0;
		  total = 0;
		  ip =
		    in + (k - 1) * x_size * y_size + (i - 1) * x_size + j - 1;
		  centre = in[k * x_size * y_size + i * x_size + j];
		  cp = bp + centre;

		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  ip += x_size - 2;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  ip += x_size - 2;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  ip += x_size * (y_size - 2) - 2;

		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  ip += x_size - 2;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  ip += x_size - 2;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  ip += x_size * (y_size - 2) - 2;

		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  ip += x_size - 2;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  ip += x_size - 2;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip++;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;
		  brightness = *ip;
		  tmp = *(cp - brightness);
		  area += tmp;
		  total += tmp * brightness;

		  if (use_median)
		    {
		      tmp = area - 100;
		      if (tmp < 200)
			*out++ = susan_median3d (in, i, j, k, x_size, y_size);
		      else
			*out++ = (total - (centre * 100)) / tmp;
		    }
		  else
		    *out++ = total / area;
		}
	    }
	}

      /* }}} */
    }
  if (tmp_image)
    delete tmp_image;
}


void
int_to_uchar (int *r, uchar * in, int size)
{
  int i, max_r = r[0], min_r = r[0];

  for (i = 0; i < size; i++)
    {
      if (r[i] > max_r)
	max_r = r[i];
      if (r[i] < min_r)
	min_r = r[i];
    }

  /*printf("min=%d max=%d\n",min_r,max_r); */

  max_r -= min_r;

  for (i = 0; i < size; i++)
    in[i] = (uchar) ((int) ((int) (r[i] - min_r) * 255) / max_r);
}

int
getint (FILE * fd)
{
  int c, i;
  char dummy[10000];

  c = getc (fd);
  while (1)			/* find next integer */
    {
      if (c == '#')		/* if we're at a comment, read to end of line */
	fgets (dummy, 9000, fd);
      if (c == EOF)
	exit_error ("Image %s not binary PGM.\n", "is");
      if (c >= '0' && c <= '9')
	break;			/* found what we were looking for */
      c = getc (fd);
    }

  /* we're at the start of a number, continue until we hit a non-number */
  i = 0;
  while (1)
    {
      i = (i * 10) + (c - '0');
      c = getc (fd);
      if (c == EOF)
	return (i);
      if (c < '0' || c > '9')
	break;
    }

  return (i);
}

void
get_image (char *filename, unsigned char **in, int *x_size, int *y_size)
{
  FILE *fd;
  char header[100];
  int tmp;

#ifdef FOPENB
  if ((fd = fopen (filename, "rb")) == NULL)
#else
  if ((fd = fopen (filename, "rb")) == NULL)
#endif
    exit_error ("Can't input image %s.\n", filename);

  /* {{{ read header */

  header[0] = fgetc (fd);
  header[1] = fgetc (fd);
  if (!(header[0] == 'P' && header[1] == '5'))
    exit_error ("Image %s does not have binary PGM header.\n", filename);

  *x_size = getint (fd);
  *y_size = getint (fd);
  tmp = getint (fd);

/* }}} */
  int readLength = (*x_size) * (*y_size);


  *in = (uchar *) malloc (readLength * sizeof (uchar));
  for (int i = 0; i < readLength; i++)
    (*in)[i] = 0;


  int val = fread (*in, 1, readLength, fd);
  printf ("READ VALUE %i \n", val);
  if (val == 0)
    exit_error ("Image %s is wrong size.\n", filename);

  //  for(int i=0;i<readLength;i++)printf(" %i ",(*in)[i]);

  fclose (fd);
}

/* }}} */
/* {{{ put_image(filename,in,x_size,y_size) */

/* }}} */
/* {{{ setup_brightness_lut(bp,thresh,form) */

void
setup_brightness_lut (uchar ** bp, int thresh, int form)
{
  int k;
  float temp;

  *bp = (unsigned char *) malloc (516);
  *bp = *bp + 258;

  for (k = -256; k < 257; k++)
    {
      temp = ((float) k) / ((float) thresh);
      temp = temp * temp;
      if (form == 6)
	temp = temp * temp * temp;
      temp = 100.0 * exp (-temp);
      *(*bp + k) = (uchar) temp;
    }
}



void
put_image (char *filename, void *in, int x_size, int y_size)
{
  FILE *fd;

#ifdef FOPENB
  if ((fd = fopen (filename, "wb")) == NULL)
#else
  if ((fd = fopen (filename, "wb")) == NULL)
#endif
    exit_error ("Can't output image%s.\n", filename);

  fprintf (fd, "P5\n");
  fprintf (fd, "%d %d\n", x_size, y_size);
  fprintf (fd, "255\n");

  if (fwrite (in, x_size * y_size, 1, fd) != 1)
    exit_error ("Can't write image %s.\n", filename);

  fclose (fd);
}

/* }}} */
/* {{{ susan principle */

/* {{{ susan_principle(in,r,bp,max_no,x_size,y_size) */

void
susan_principle (uchar * in, int *r, uchar * bp, int max_no, int x_size,
		 int y_size)
{

  int i, j, n;
  uchar *p, *cp;

  memset (r, 0, x_size * y_size * sizeof (int));

  for (i = 3; i < y_size - 3; i++)
    for (j = 3; j < x_size - 3; j++)
      {
	n = 100;
	p = in + (i - 3) * x_size + j - 1;
	cp = bp + in[i * x_size + j];

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 3;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 5;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 6;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += 2;
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 6;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 5;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 3;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);

	if (n <= max_no)
	  r[i * x_size + j] = max_no - n;
      }
}

/* }}} */
/* {{{ susan_principle_small(in,r,bp,max_no,x_size,y_size) */

void
susan_principle_small (uchar * in, int *r, uchar * bp, int max_no, int x_size,
		       int y_size)
{
  int i, j, n;
  uchar *p, *cp;

  memset (r, 0, x_size * y_size * sizeof (int));

  max_no = 730;			/* ho hum ;) */

  for (i = 1; i < y_size - 1; i++)
    for (j = 1; j < x_size - 1; j++)
      {
	n = 100;
	p = in + (i - 1) * x_size + j - 1;
	cp = bp + in[i * x_size + j];

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 2;

	n += *(cp - *p);
	p += 2;
	n += *(cp - *p);
	p += x_size - 2;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);

	if (n <= max_no)
	  r[i * x_size + j] = max_no - n;
      }
}

/* }}} */

/* }}} */
/* {{{ smoothing */

/* {{{ median(in,i,j,x_size) */

uchar
median (uchar * in, int i, int j, int x_size)
{
  int p[8], k, l, tmp;

  p[0] = in[(i - 1) * x_size + j - 1];
  p[1] = in[(i - 1) * x_size + j];
  p[2] = in[(i - 1) * x_size + j + 1];
  p[3] = in[(i) * x_size + j - 1];
  p[4] = in[(i) * x_size + j + 1];
  p[5] = in[(i + 1) * x_size + j - 1];
  p[6] = in[(i + 1) * x_size + j];
  p[7] = in[(i + 1) * x_size + j + 1];

  for (k = 0; k < 7; k++)
    for (l = 0; l < (7 - k); l++)
      if (p[l] > p[l + 1])
	{
	  tmp = p[l];
	  p[l] = p[l + 1];
	  p[l + 1] = tmp;
	}

  return ((p[3] + p[4]) / 2);
}

/* }}} */
/* {{{ enlarge(in,tmp_image,x_size,y_size,border) */

/* this enlarges "in" so that borders can be dealt with easily */

void
enlarge (uchar ** in, uchar * tmp_image, int *x_size, int *y_size, int border)
{
  int i, j;

  for (i = 0; i < *y_size; i++)	/* copy *in into tmp_image */
    memcpy (tmp_image + (i + border) * (*x_size + 2 * border) + border,
	    *in + i * *x_size, *x_size);

  for (i = 0; i < border; i++)	/* copy top and bottom rows; invert as many as necessary */
    {
      memcpy (tmp_image + (border - 1 - i) * (*x_size + 2 * border) + border,
	      *in + i * *x_size, *x_size);
      memcpy (tmp_image + (*y_size + border + i) * (*x_size + 2 * border) +
	      border, *in + (*y_size - i - 1) * *x_size, *x_size);
    }

  for (i = 0; i < border; i++)	/* copy left and right columns */
    for (j = 0; j < *y_size + 2 * border; j++)
      {
	tmp_image[j * (*x_size + 2 * border) + border - 1 - i] =
	  tmp_image[j * (*x_size + 2 * border) + border + i];
	tmp_image[j * (*x_size + 2 * border) + *x_size + border + i] =
	  tmp_image[j * (*x_size + 2 * border) + *x_size + border - 1 - i];
      }

  *x_size += 2 * border;	/* alter image size */
  *y_size += 2 * border;
  *in = tmp_image;		/* repoint in */
}

/* }}} */
/* {{{ void susan_smoothing(three_by_three,in,dt,x_size,y_size,bp) */

void
susan_smoothing (int three_by_three, uchar * in, float dt, int x_size,
		 int y_size, uchar * bp)
{
  /* {{{ vars */

  float temp;
  int n_max, increment, mask_size, i, j, x, y, area, brightness, tmp, centre;
  uchar *ip, *dp, *dpt, *cp, *out = in, *tmp_image;
  TOTAL_TYPE total = 0; 

  /* }}} */

  /* {{{ setup larger image and border sizes */

  if (three_by_three == 0)
    mask_size = ((int) (1.5 * dt)) + 1;
  else
    mask_size = 1;

  if ((dt > 15) && (total == 0))
    {
      printf ("Distance_thresh (%f) too big for integer arithmetic.\n", dt);
      printf
	("Either reduce it to <=15 or recompile with variable \"total\"\n");
      printf ("as a float: see top \"defines\" section.\n");
      exit (0);
    }

  if ((2 * mask_size + 1 > x_size) || (2 * mask_size + 1 > y_size))
    {
      printf
	("Mask size (1.5*distance_thresh+1=%d) too big for image (%dx%d).\n",
	 mask_size, x_size, y_size);
      exit (0);
    }

  tmp_image =
    (uchar *) malloc ((x_size + mask_size * 2) * (y_size + mask_size * 2));
  enlarge (&in, tmp_image, &x_size, &y_size, mask_size);

  /* }}} */

  if (three_by_three == 0)
    {				/* large Gaussian masks */
      /* {{{ setup distance lut */

      n_max = (mask_size * 2) + 1;

      increment = x_size - n_max;

      dp = (unsigned char *) malloc (n_max * n_max);
      dpt = dp;
      temp = -(dt * dt);

      for (i = -mask_size; i <= mask_size; i++)
	for (j = -mask_size; j <= mask_size; j++)
	  {
	    x = (int) (100.0 * exp (((float) ((i * i) + (j * j))) / temp));
	    *dpt++ = (unsigned char) x;
	  }

      /* }}} */
      /* {{{ main section */

      for (i = mask_size; i < y_size - mask_size; i++)
	{
	  for (j = mask_size; j < x_size - mask_size; j++)
	    {
	      area = 0;
	      total = 0;
	      dpt = dp;
	      ip = in + ((i - mask_size) * x_size) + j - mask_size;
	      centre = in[i * x_size + j];
	      cp = bp + centre;
	      for (y = -mask_size; y <= mask_size; y++)
		{
		  for (x = -mask_size; x <= mask_size; x++)
		    {
		      brightness = *ip++;
		      tmp = *dpt++ * *(cp - brightness);
		      area += tmp;
		      total += tmp * brightness;
		    }
		  ip += increment;
		}
	      tmp = area - 10000;
	      if (tmp == 0)
		*out++ = median (in, i, j, x_size);
	      else
		*out++ = ((total - (centre * 10000)) / tmp);
	    }
	}

      /* }}} */
    }
  else
    {				/* 3x3 constant mask */
      /* {{{ main section */

      for (i = 1; i < y_size - 1; i++)
	{
	  for (j = 1; j < x_size - 1; j++)
	    {
	      area = 0;
	      total = 0;
	      ip = in + ((i - 1) * x_size) + j - 1;
	      centre = in[i * x_size + j];
	      cp = bp + centre;

	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      ip += x_size - 2;
	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      ip += x_size - 2;
	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip++;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;
	      brightness = *ip;
	      tmp = *(cp - brightness);
	      area += tmp;
	      total += tmp * brightness;

	      tmp = area - 100;
	      if (tmp == 0)
		*out++ = median (in, i, j, x_size);
	      else
		*out++ = (total - (centre * 100)) / tmp;
	    }
	}

      /* }}} */
    }
  if (tmp_image)
    delete tmp_image;
}

/* }}} */

/* }}} */
/* {{{ edges */

/* {{{ edge_draw(in,corner_list,drawing_mode) */

void
edge_draw (uchar * in, uchar * mid, int x_size, int y_size, int drawing_mode)
{
  int i;
  uchar *inp, *midp;

  if (drawing_mode == 0)
    {
      /* mark 3x3 white block around each edge point */
      midp = mid;
      for (i = 0; i < x_size * y_size; i++)
	{
	  if (*midp < 8)
	    {
	      inp = in + (midp - mid) - x_size - 1;
	      *inp++ = 255;
	      *inp++ = 255;
	      *inp = 255;
	      inp += x_size - 2;
	      *inp++ = 255;
	      *inp++;
	      *inp = 255;
	      inp += x_size - 2;
	      *inp++ = 255;
	      *inp++ = 255;
	      *inp = 255;
	    }
	  midp++;
	}
    }

  /* now mark 1 black pixel at each edge point */
  midp = mid;
  for (i = 0; i < x_size * y_size; i++)
    {
      if (*midp < 8)
	*(in + (midp - mid)) = 0;
      midp++;
    }
}

/* }}} */
/* {{{ susan_thin(r,mid,x_size,y_size) */

/* only one pass is needed as i,j are decremented if necessary to go
back and do bits again */

void
susan_thin (int *r, uchar * mid, int x_size, int y_size)
{
  int l[9], centre,
    b01, b12, b21, b10,
    p1, p2, p3, p4, b00, b02, b20, b22, m, n, a, b, x, y, i, j;
  uchar *mp;
  a=b=0;

  for (i = 4; i < y_size - 4; i++)
    for (j = 4; j < x_size - 4; j++)
      if (mid[i * x_size + j] < 8)
	{
	  centre = r[i * x_size + j];
	  /* {{{ count number of neighbours */

	  mp = mid + (i - 1) * x_size + j - 1;

	  n = (*mp < 8) +
	    (*(mp + 1) < 8) +
	    (*(mp + 2) < 8) +
	    (*(mp + x_size) < 8) +
	    (*(mp + x_size + 2) < 8) +
	    (*(mp + x_size + x_size) < 8) +
	    (*(mp + x_size + x_size + 1) < 8) +
	    (*(mp + x_size + x_size + 2) < 8);

	  /* }}} */
	  /* {{{ n==0 no neighbours - remove point */

	  if (n == 0)
	    mid[i * x_size + j] = 100;

	  /* }}} */
	  /* {{{ n==1 - extend line if I can */

	  /* extension is only allowed a few times - the value of mid is used to control this */

	  if ((n == 1) && (mid[i * x_size + j] < 6))
	    {
	      /* find maximum neighbour weighted in direction opposite the
	         neighbour already present. e.g.
	         have: O O O  weight r by 0 2 3
	         X X O              0 0 4
	         O O O              0 2 3     */

	      l[0] = r[(i - 1) * x_size + j - 1];
	      l[1] = r[(i - 1) * x_size + j];
	      l[2] = r[(i - 1) * x_size + j + 1];
	      l[3] = r[(i) * x_size + j - 1];
	      l[4] = 0;
	      l[5] = r[(i) * x_size + j + 1];
	      l[6] = r[(i + 1) * x_size + j - 1];
	      l[7] = r[(i + 1) * x_size + j];
	      l[8] = r[(i + 1) * x_size + j + 1];

	      if (mid[(i - 1) * x_size + j - 1] < 8)
		{
		  l[0] = 0;
		  l[1] = 0;
		  l[3] = 0;
		  l[2] *= 2;
		  l[6] *= 2;
		  l[5] *= 3;
		  l[7] *= 3;
		  l[8] *= 4;
		}
	      else
		{
		  if (mid[(i - 1) * x_size + j] < 8)
		    {
		      l[1] = 0;
		      l[0] = 0;
		      l[2] = 0;
		      l[3] *= 2;
		      l[5] *= 2;
		      l[6] *= 3;
		      l[8] *= 3;
		      l[7] *= 4;
		    }
		  else
		    {
		      if (mid[(i - 1) * x_size + j + 1] < 8)
			{
			  l[2] = 0;
			  l[1] = 0;
			  l[5] = 0;
			  l[0] *= 2;
			  l[8] *= 2;
			  l[3] *= 3;
			  l[7] *= 3;
			  l[6] *= 4;
			}
		      else
			{
			  if (mid[(i) * x_size + j - 1] < 8)
			    {
			      l[3] = 0;
			      l[0] = 0;
			      l[6] = 0;
			      l[1] *= 2;
			      l[7] *= 2;
			      l[2] *= 3;
			      l[8] *= 3;
			      l[5] *= 4;
			    }
			  else
			    {
			      if (mid[(i) * x_size + j + 1] < 8)
				{
				  l[5] = 0;
				  l[2] = 0;
				  l[8] = 0;
				  l[1] *= 2;
				  l[7] *= 2;
				  l[0] *= 3;
				  l[6] *= 3;
				  l[3] *= 4;
				}
			      else
				{
				  if (mid[(i + 1) * x_size + j - 1] < 8)
				    {
				      l[6] = 0;
				      l[3] = 0;
				      l[7] = 0;
				      l[0] *= 2;
				      l[8] *= 2;
				      l[1] *= 3;
				      l[5] *= 3;
				      l[2] *= 4;
				    }
				  else
				    {
				      if (mid[(i + 1) * x_size + j] < 8)
					{
					  l[7] = 0;
					  l[6] = 0;
					  l[8] = 0;
					  l[3] *= 2;
					  l[5] *= 2;
					  l[0] *= 3;
					  l[2] *= 3;
					  l[1] *= 4;
					}
				      else
					{
					  if (mid[(i + 1) * x_size + j + 1] <
					      8)
					    {
					      l[8] = 0;
					      l[5] = 0;
					      l[7] = 0;
					      l[6] *= 2;
					      l[2] *= 2;
					      l[1] *= 3;
					      l[3] *= 3;
					      l[0] *= 4;
					    }
					}
				    }
				}
			    }
			}
		    }
		}

	      m = 0;		/* find the highest point */
	      for (y = 0; y < 3; y++)
		for (x = 0; x < 3; x++)
		  if (l[y + y + y + x] > m)
		    {
		      m = l[y + y + y + x];
		      a = y;
		      b = x;
		    }

	      if (m > 0)
		{
		  if (mid[i * x_size + j] < 4)
		    mid[(i + a - 1) * x_size + j + b - 1] = 4;
		  else
		    mid[(i + a - 1) * x_size + j + b - 1] =
		      mid[i * x_size + j] + 1;
		  if ((a + a + b) < 3)	/* need to jump back in image */
		    {
		      i += a - 1;
		      j += b - 2;
		      if (i < 4)
			i = 4;
		      if (j < 4)
			j = 4;
		    }
		}
	    }

	  /* }}} */
	  /* {{{ n==2 */

	  if (n == 2)
	    {
	      /* put in a bit here to straighten edges */
	      b00 = mid[(i - 1) * x_size + j - 1] < 8;	/* corners of 3x3 */
	      b02 = mid[(i - 1) * x_size + j + 1] < 8;
	      b20 = mid[(i + 1) * x_size + j - 1] < 8;
	      b22 = mid[(i + 1) * x_size + j + 1] < 8;
	      if (((b00 + b02 + b20 + b22) == 2)
		  && ((b00 | b22) & (b02 | b20)))
		{		/* case: move a point back into line.
				   e.g. X O X  CAN  become X X X
				   O X O              O O O
				   O O O              O O O    */
		  if (b00)
		    {
		      if (b02)
			{
			  x = 0;
			  y = -1;
			}
		      else
			{
			  x = -1;
			  y = 0;
			}
		    }
		  else
		    {
		      if (b02)
			{
			  x = 1;
			  y = 0;
			}
		      else
			{
			  x = 0;
			  y = 1;
			}
		    }
		  if (((float) r[(i + y) * x_size + j + x] / (float) centre) >
		      0.7)
		    {
		      if (((x == 0) && (mid[(i + (2 * y)) * x_size + j] > 7)
			   && (mid[(i + (2 * y)) * x_size + j - 1] > 7)
			   && (mid[(i + (2 * y)) * x_size + j + 1] > 7))
			  || ((y == 0)
			      && (mid[(i) * x_size + j + (2 * x)] > 7)
			      && (mid[(i + 1) * x_size + j + (2 * x)] > 7)
			      && (mid[(i - 1) * x_size + j + (2 * x)] > 7)))
			{
			  mid[(i) * x_size + j] = 100;
			  mid[(i + y) * x_size + j + x] = 3;	/* no jumping needed */
			}
		    }
		}
	      else
		{
		  b01 = mid[(i - 1) * x_size + j] < 8;
		  b12 = mid[(i) * x_size + j + 1] < 8;
		  b21 = mid[(i + 1) * x_size + j] < 8;
		  b10 = mid[(i) * x_size + j - 1] < 8;
		  /* {{{ right angle ends - not currently used */

#ifdef IGNORETHIS
		  if ((b00 & b01) | (b00 & b10) | (b02 & b01) | (b02 & b12) |
		      (b20 & b10) | (b20 & b21) | (b22 & b21) | (b22 & b12))
		    {		/* case; right angle ends. clean up.
				   e.g.; X X O  CAN  become X X O
				   O X O              O O O
				   O O O              O O O        */
		      if (((b01) & (mid[(i - 2) * x_size + j - 1] > 7) &
			   (mid[(i - 2) * x_size + j] >
			    7) & (mid[(i - 2) * x_size + j + 1] >
				  7) & ((b00 &
					 ((2 * r[(i - 1) * x_size + j + 1]) >
					  centre)) | (b02 & ((2 *
							      r[(i -
								 1) * x_size +
								j - 1]) >
							     centre)))) |
			  ((b10) & (mid[(i - 1) * x_size + j - 2] > 7) &
			   (mid[(i) * x_size + j - 2] >
			    7) & (mid[(i + 1) * x_size + j - 2] >
				  7) & ((b00 &
					 ((2 * r[(i + 1) * x_size + j - 1]) >
					  centre)) | (b20 & ((2 *
							      r[(i -
								 1) * x_size +
								j - 1]) >
							     centre)))) |
			  ((b12) & (mid[(i - 1) * x_size + j + 2] > 7) &
			   (mid[(i) * x_size + j + 2] >
			    7) & (mid[(i + 1) * x_size + j + 2] >
				  7) & ((b02 &
					 ((2 * r[(i + 1) * x_size + j + 1]) >
					  centre)) | (b22 & ((2 *
							      r[(i -
								 1) * x_size +
								j + 1]) >
							     centre)))) |
			  ((b21) & (mid[(i + 2) * x_size + j - 1] > 7) &
			   (mid[(i + 2) * x_size + j] >
			    7) & (mid[(i + 2) * x_size + j + 1] >
				  7) & ((b20 &
					 ((2 * r[(i + 1) * x_size + j + 1]) >
					  centre)) | (b22 & ((2 *
							      r[(i +
								 1) * x_size +
								j - 1]) >
							     centre)))))
			{
			  mid[(i) * x_size + j] = 100;
			  if (b10 & b20)
			    j -= 2;
			  if (b00 | b01 | b02)
			    {
			      i--;
			      j -= 2;
			    }
			}
		    }
#endif

		  /* }}} */
		  if (((b01 + b12 + b21 + b10) == 2)
		      && ((b10 | b12) & (b01 | b21))
		      &&
		      ((b01 &
			((mid[(i - 2) * x_size + j - 1] <
			  8) | (mid[(i - 2) * x_size + j + 1] <
				8))) | (b10 & ((mid[(i - 1) * x_size + j -
						    2] <
						8) | (mid[(i + 1) * x_size +
							  j - 2] <
						      8))) | (b12 &
							      ((mid
								[(i -
								  1) *
								 x_size + j +
								 2] <
								8) |
							       (mid
								[(i +
								  1) *
								 x_size + j +
								 2] <
								8))) | (b21 &
									((mid
									  [(i
									    +
									    2)
									   *
									   x_size
									   +
									   j -
									   1]
									  <
									  8) |
									 (mid
									  [(i
									    +
									    2)
									   *
									   x_size
									   +
									   j +
									   1]
									  <
									  8)))))
		    {		/* case; clears odd right angles.
				   e.g.; O O O  becomes O O O
				   X X O          X O O
				   O X O          O X O     */
		      mid[(i) * x_size + j] = 100;
		      i--;	/* jump back */
		      j -= 2;
		      if (i < 4)
			i = 4;
		      if (j < 4)
			j = 4;
		    }
		}
	    }

	  /* }}} */
	  /* {{{ n>2 the thinning is done here without breaking connectivity */

	  if (n > 2)
	    {
	      b01 = mid[(i - 1) * x_size + j] < 8;
	      b12 = mid[(i) * x_size + j + 1] < 8;
	      b21 = mid[(i + 1) * x_size + j] < 8;
	      b10 = mid[(i) * x_size + j - 1] < 8;
	      if ((b01 + b12 + b21 + b10) > 1)
		{
		  b00 = mid[(i - 1) * x_size + j - 1] < 8;
		  b02 = mid[(i - 1) * x_size + j + 1] < 8;
		  b20 = mid[(i + 1) * x_size + j - 1] < 8;
		  b22 = mid[(i + 1) * x_size + j + 1] < 8;
		  p1 = b00 | b01;
		  p2 = b02 | b12;
		  p3 = b22 | b21;
		  p4 = b20 | b10;

		  if (((p1 + p2 + p3 + p4) -
		       ((b01 & p2) + (b12 & p3) + (b21 & p4) + (b10 & p1))) <
		      2)
		    {
		      mid[(i) * x_size + j] = 100;
		      i--;
		      j -= 2;
		      if (i < 4)
			i = 4;
		      if (j < 4)
			j = 4;
		    }
		}
	    }

	  /* }}} */
	}
}

/* }}} */
/* {{{ susan_edges(in,r,sf,max_no,out) */

void
susan_edges (uchar * in, int *r, uchar * mid, uchar * bp, int max_no,
	     int x_size, int y_size)
{
  float z;
  int do_symmetry, i, j, m, n, a, b, x, y, w;
  uchar c, *p, *cp;

  memset (r, 0, x_size * y_size * sizeof (int));

  for (i = 3; i < y_size - 3; i++)
    for (j = 3; j < x_size - 3; j++)
      {
	n = 100;
	p = in + (i - 3) * x_size + j - 1;
	cp = bp + in[i * x_size + j];

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 3;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 5;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 6;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += 2;
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 6;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 5;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 3;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);

	if (n <= max_no)
	  r[i * x_size + j] = max_no - n;
      }

  for (i = 4; i < y_size - 4; i++)
    for (j = 4; j < x_size - 4; j++)
      {
	if (r[i * x_size + j] > 0)
	  {
	    m = r[i * x_size + j];
	    n = max_no - m;
	    cp = bp + in[i * x_size + j];

	    if (n > 600)
	      {
		p = in + (i - 3) * x_size + j - 1;
		x = 0;
		y = 0;

		c = *(cp - *p++);
		x -= c;
		y -= 3 * c;
		c = *(cp - *p++);
		y -= 3 * c;
		c = *(cp - *p);
		x += c;
		y -= 3 * c;
		p += x_size - 3;

		c = *(cp - *p++);
		x -= 2 * c;
		y -= 2 * c;
		c = *(cp - *p++);
		x -= c;
		y -= 2 * c;
		c = *(cp - *p++);
		y -= 2 * c;
		c = *(cp - *p++);
		x += c;
		y -= 2 * c;
		c = *(cp - *p);
		x += 2 * c;
		y -= 2 * c;
		p += x_size - 5;

		c = *(cp - *p++);
		x -= 3 * c;
		y -= c;
		c = *(cp - *p++);
		x -= 2 * c;
		y -= c;
		c = *(cp - *p++);
		x -= c;
		y -= c;
		c = *(cp - *p++);
		y -= c;
		c = *(cp - *p++);
		x += c;
		y -= c;
		c = *(cp - *p++);
		x += 2 * c;
		y -= c;
		c = *(cp - *p);
		x += 3 * c;
		y -= c;
		p += x_size - 6;

		c = *(cp - *p++);
		x -= 3 * c;
		c = *(cp - *p++);
		x -= 2 * c;
		c = *(cp - *p);
		x -= c;
		p += 2;
		c = *(cp - *p++);
		x += c;
		c = *(cp - *p++);
		x += 2 * c;
		c = *(cp - *p);
		x += 3 * c;
		p += x_size - 6;

		c = *(cp - *p++);
		x -= 3 * c;
		y += c;
		c = *(cp - *p++);
		x -= 2 * c;
		y += c;
		c = *(cp - *p++);
		x -= c;
		y += c;
		c = *(cp - *p++);
		y += c;
		c = *(cp - *p++);
		x += c;
		y += c;
		c = *(cp - *p++);
		x += 2 * c;
		y += c;
		c = *(cp - *p);
		x += 3 * c;
		y += c;
		p += x_size - 5;

		c = *(cp - *p++);
		x -= 2 * c;
		y += 2 * c;
		c = *(cp - *p++);
		x -= c;
		y += 2 * c;
		c = *(cp - *p++);
		y += 2 * c;
		c = *(cp - *p++);
		x += c;
		y += 2 * c;
		c = *(cp - *p);
		x += 2 * c;
		y += 2 * c;
		p += x_size - 3;

		c = *(cp - *p++);
		x -= c;
		y += 3 * c;
		c = *(cp - *p++);
		y += 3 * c;
		c = *(cp - *p);
		x += c;
		y += 3 * c;

		z = sqrt ((float) ((x * x) + (y * y)));
		if (z > (0.9 * (float) n))	/* 0.5 */
		  {
		    do_symmetry = 0;
		    if (x == 0)
		      z = 1000000.0;
		    else
		      z = ((float) y) / ((float) x);
		    if (z < 0)
		      {
			z = -z;
			w = -1;
		      }
		    else
		      w = 1;
		    if (z < 0.5)
		      {		/* vert_edge */
			a = 0;
			b = 1;
		      }
		    else
		      {
			if (z > 2.0)
			  {	/* hor_edge */
			    a = 1;
			    b = 0;
			  }
			else
			  {	/* diag_edge */
			    if (w > 0)
			      {
				a = 1;
				b = 1;
			      }
			    else
			      {
				a = -1;
				b = 1;
			      }
			  }
		      }
		    if ((m > r[(i + a) * x_size + j + b])
			&& (m >= r[(i - a) * x_size + j - b])
			&& (m > r[(i + (2 * a)) * x_size + j + (2 * b)])
			&& (m >= r[(i - (2 * a)) * x_size + j - (2 * b)]))
		      mid[i * x_size + j] = 1;
		  }
		else
		  do_symmetry = 1;
	      }
	    else
	      do_symmetry = 1;

	    if (do_symmetry == 1)
	      {
		p = in + (i - 3) * x_size + j - 1;
		x = 0;
		y = 0;
		w = 0;

		/*   |      \
		   y  -x-  w
		   |        \   */

		c = *(cp - *p++);
		x += c;
		y += 9 * c;
		w += 3 * c;
		c = *(cp - *p++);
		y += 9 * c;
		c = *(cp - *p);
		x += c;
		y += 9 * c;
		w -= 3 * c;
		p += x_size - 3;

		c = *(cp - *p++);
		x += 4 * c;
		y += 4 * c;
		w += 4 * c;
		c = *(cp - *p++);
		x += c;
		y += 4 * c;
		w += 2 * c;
		c = *(cp - *p++);
		y += 4 * c;
		c = *(cp - *p++);
		x += c;
		y += 4 * c;
		w -= 2 * c;
		c = *(cp - *p);
		x += 4 * c;
		y += 4 * c;
		w -= 4 * c;
		p += x_size - 5;

		c = *(cp - *p++);
		x += 9 * c;
		y += c;
		w += 3 * c;
		c = *(cp - *p++);
		x += 4 * c;
		y += c;
		w += 2 * c;
		c = *(cp - *p++);
		x += c;
		y += c;
		w += c;
		c = *(cp - *p++);
		y += c;
		c = *(cp - *p++);
		x += c;
		y += c;
		w -= c;
		c = *(cp - *p++);
		x += 4 * c;
		y += c;
		w -= 2 * c;
		c = *(cp - *p);
		x += 9 * c;
		y += c;
		w -= 3 * c;
		p += x_size - 6;

		c = *(cp - *p++);
		x += 9 * c;
		c = *(cp - *p++);
		x += 4 * c;
		c = *(cp - *p);
		x += c;
		p += 2;
		c = *(cp - *p++);
		x += c;
		c = *(cp - *p++);
		x += 4 * c;
		c = *(cp - *p);
		x += 9 * c;
		p += x_size - 6;

		c = *(cp - *p++);
		x += 9 * c;
		y += c;
		w -= 3 * c;
		c = *(cp - *p++);
		x += 4 * c;
		y += c;
		w -= 2 * c;
		c = *(cp - *p++);
		x += c;
		y += c;
		w -= c;
		c = *(cp - *p++);
		y += c;
		c = *(cp - *p++);
		x += c;
		y += c;
		w += c;
		c = *(cp - *p++);
		x += 4 * c;
		y += c;
		w += 2 * c;
		c = *(cp - *p);
		x += 9 * c;
		y += c;
		w += 3 * c;
		p += x_size - 5;

		c = *(cp - *p++);
		x += 4 * c;
		y += 4 * c;
		w -= 4 * c;
		c = *(cp - *p++);
		x += c;
		y += 4 * c;
		w -= 2 * c;
		c = *(cp - *p++);
		y += 4 * c;
		c = *(cp - *p++);
		x += c;
		y += 4 * c;
		w += 2 * c;
		c = *(cp - *p);
		x += 4 * c;
		y += 4 * c;
		w += 4 * c;
		p += x_size - 3;

		c = *(cp - *p++);
		x += c;
		y += 9 * c;
		w -= 3 * c;
		c = *(cp - *p++);
		y += 9 * c;
		c = *(cp - *p);
		x += c;
		y += 9 * c;
		w += 3 * c;

		if (y == 0)
		  z = 1000000.0;
		else
		  z = ((float) x) / ((float) y);
		if (z < 0.5)
		  {		/* vertical */
		    a = 0;
		    b = 1;
		  }
		else
		  {
		    if (z > 2.0)
		      {		/* horizontal */
			a = 1;
			b = 0;
		      }
		    else
		      {		/* diagonal */
			if (w > 0)
			  {
			    a = -1;
			    b = 1;
			  }
			else
			  {
			    a = 1;
			    b = 1;
			  }
		      }
		  }
		if ((m > r[(i + a) * x_size + j + b])
		    && (m >= r[(i - a) * x_size + j - b])
		    && (m > r[(i + (2 * a)) * x_size + j + (2 * b)])
		    && (m >= r[(i - (2 * a)) * x_size + j - (2 * b)]))
		  mid[i * x_size + j] = 2;
	      }
	  }
      }
}

/* }}} */
/* {{{ susan_edges_small(in,r,sf,max_no,out) */

void
susan_edges_small (uchar * in, int *r, uchar * mid, uchar * bp, int max_no,
		   int x_size, int y_size)
{
  float z;
  int do_symmetry, i, j, m, n, a, b, x, y, w;
  uchar c, *p, *cp;

  memset (r, 0, x_size * y_size * sizeof (int));

  max_no = 730;			/* ho hum ;) */

  for (i = 1; i < y_size - 1; i++)
    for (j = 1; j < x_size - 1; j++)
      {
	n = 100;
	p = in + (i - 1) * x_size + j - 1;
	cp = bp + in[i * x_size + j];

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);
	p += x_size - 2;

	n += *(cp - *p);
	p += 2;
	n += *(cp - *p);
	p += x_size - 2;

	n += *(cp - *p++);
	n += *(cp - *p++);
	n += *(cp - *p);

	if (n <= max_no)
	  r[i * x_size + j] = max_no - n;
      }

  for (i = 2; i < y_size - 2; i++)
    for (j = 2; j < x_size - 2; j++)
      {
	if (r[i * x_size + j] > 0)
	  {
	    m = r[i * x_size + j];
	    n = max_no - m;
	    cp = bp + in[i * x_size + j];

	    if (n > 250)
	      {
		p = in + (i - 1) * x_size + j - 1;
		x = 0;
		y = 0;

		c = *(cp - *p++);
		x -= c;
		y -= c;
		c = *(cp - *p++);
		y -= c;
		c = *(cp - *p);
		x += c;
		y -= c;
		p += x_size - 2;

		c = *(cp - *p);
		x -= c;
		p += 2;
		c = *(cp - *p);
		x += c;
		p += x_size - 2;

		c = *(cp - *p++);
		x -= c;
		y += c;
		c = *(cp - *p++);
		y += c;
		c = *(cp - *p);
		x += c;
		y += c;

		z = sqrt ((float) ((x * x) + (y * y)));
		if (z > (0.4 * (float) n))	/* 0.6 */
		  {
		    do_symmetry = 0;
		    if (x == 0)
		      z = 1000000.0;
		    else
		      z = ((float) y) / ((float) x);
		    if (z < 0)
		      {
			z = -z;
			w = -1;
		      }
		    else
		      w = 1;
		    if (z < 0.5)
		      {		/* vert_edge */
			a = 0;
			b = 1;
		      }
		    else
		      {
			if (z > 2.0)
			  {	/* hor_edge */
			    a = 1;
			    b = 0;
			  }
			else
			  {	/* diag_edge */
			    if (w > 0)
			      {
				a = 1;
				b = 1;
			      }
			    else
			      {
				a = -1;
				b = 1;
			      }
			  }
		      }
		    if ((m > r[(i + a) * x_size + j + b])
			&& (m >= r[(i - a) * x_size + j - b]))
		      mid[i * x_size + j] = 1;
		  }
		else
		  do_symmetry = 1;
	      }
	    else
	      do_symmetry = 1;

	    if (do_symmetry == 1)
	      {
		p = in + (i - 1) * x_size + j - 1;
		x = 0;
		y = 0;
		w = 0;

		/*   |      \
		   y  -x-  w
		   |        \   */

		c = *(cp - *p++);
		x += c;
		y += c;
		w += c;
		c = *(cp - *p++);
		y += c;
		c = *(cp - *p);
		x += c;
		y += c;
		w -= c;
		p += x_size - 2;

		c = *(cp - *p);
		x += c;
		p += 2;
		c = *(cp - *p);
		x += c;
		p += x_size - 2;

		c = *(cp - *p++);
		x += c;
		y += c;
		w -= c;
		c = *(cp - *p++);
		y += c;
		c = *(cp - *p);
		x += c;
		y += c;
		w += c;

		if (y == 0)
		  z = 1000000.0;
		else
		  z = ((float) x) / ((float) y);
		if (z < 0.5)
		  {		/* vertical */
		    a = 0;
		    b = 1;
		  }
		else
		  {
		    if (z > 2.0)
		      {		/* horizontal */
			a = 1;
			b = 0;
		      }
		    else
		      {		/* diagonal */
			if (w > 0)
			  {
			    a = -1;
			    b = 1;
			  }
			else
			  {
			    a = 1;
			    b = 1;
			  }
		      }
		  }
		if ((m > r[(i + a) * x_size + j + b])
		    && (m >= r[(i - a) * x_size + j - b]))
		  mid[i * x_size + j] = 2;
	      }
	  }
      }
}

/* }}} */

void JNICALL Java_SusanFilter3D_susan_1smoothing3d
  (JNIEnv * env, jclass susan, jboolean use_three_by_three, jbyteArray in_arr,
   jfloat sigma, jint bt, jint x_size, jint y_size, jint z_size,
   jboolean use_median)
{

  jbyte *in = (env)->GetByteArrayElements (in_arr, JNI_FALSE);

  unsigned char *bp;

  unsigned char *in2 = (unsigned char *) in;

  setup_brightness_lut (&bp, bt, 2);

  susan_smoothing3d (env, susan, use_three_by_three, in2, sigma, sigma, sigma,
		     x_size, y_size, z_size, bp, use_median);

  (env)->ReleaseByteArrayElements (in_arr, in, 0);
}


void JNICALL Java_SusanFilter3D_susan_1smoothing2d
  (JNIEnv * env, jclass susan, jboolean use_three_by_three, jbyteArray in_arr,
   jfloat sigma, jint bt, jint x_size, jint y_size, jint z_size,
   jboolean use_median)
{

  jbyte *in = (env)->GetByteArrayElements (in_arr, JNI_FALSE);

  unsigned char *bp;

  unsigned char *in2 = (unsigned char *) in;

  setup_brightness_lut (&bp, bt, 2);
  for (int z = 0; z < z_size; z++)
    {

      susan_smoothing (use_three_by_three, &(in2[x_size * y_size * z]), sigma,
		       sigma, x_size, y_size, z_size, bp, use_median);
      setProgress (env, susan, z, z_size);
    }
  (env)->ReleaseByteArrayElements (in_arr, in, 0);
}


void JNICALL
Java_SusanFilter3D_susan_1edges (JNIEnv * env, jclass susan,
				 jboolean use_three_by_three,
				 jbyteArray in_arr, jboolean principle,
				 jboolean thin_post_proc, jint max_no_edges,
				 jint bt, jint x_size, jint y_size,
				 jint drawing_mode)
{



  jbyte *in = (env)->GetByteArrayElements (in_arr, JNI_FALSE);

  unsigned char *bp;
  unsigned char *in2 = (unsigned char *) in;
  int *r = NULL;
  uchar *mid = NULL;

  r = (int *) malloc (x_size * y_size * sizeof (int));
  if (!r)
    {
      printf ("Not enough memory! \n");
      return;
    }
  setup_brightness_lut (&bp, bt, 6);

  if (principle)
    {
      if (use_three_by_three)
	{
	  susan_principle_small (in2, r, bp, max_no_edges, x_size, y_size);
	}
      else
	{
	  susan_principle (in2, r, bp, max_no_edges, x_size, y_size);
	}
      int_to_uchar (r, in2, x_size * y_size);
    }
  else
    {
      mid = (uchar *) malloc (x_size * y_size);
      memset (mid, 100, x_size * y_size);	/* note not set to zero */

      if (use_three_by_three)
	{
	  susan_edges_small (in2, r, mid, bp, max_no_edges, x_size, y_size);
	}
      else
	{
	  susan_edges (in2, r, mid, bp, max_no_edges, x_size, y_size);
	}
      if (thin_post_proc)
	{
	  susan_thin (r, mid, x_size, y_size);
	}

      edge_draw (in2, mid, x_size, y_size, drawing_mode);
    }

//clean up
  if (mid)
    delete mid;
  if (r)
    delete r;
  (env)->ReleaseByteArrayElements (in_arr, in, 0);
}
